

set -ex



pip check
jupyter server extension list
jupyter labextension list
jupyter notebook extension list
jupyter lab extension list
jupyter server extension list 1>server_extensions 2>&1
cat server_extensions | grep -ie "jupyter_server_proxy.*OK"
jupyter labextension list 1>labextensions 2>&1
cat labextensions | grep -ie "@jupyterhub/jupyter-server-proxy.*4\.1\.2.*OK"
jupyter notebook extension list 1>notebook_extensions 2>&1
cat notebook_extensions | grep -ie "jupyter_server_proxy.*OK"
jupyter lab extension list 1>lab_extensions 2>&1
cat lab_extensions | grep -ie "jupyter_server_proxy.*OK"
exit 0
